@@ -18,32 +18,115 @@ use crate::api::{
 };
 use api::{auth::*, chat::EventSource, Hmc, HmcFromStrError};
 use error::*;
+use tracing::Span;
 
 use std::{
     convert::TryFrom,
     fmt::{self, Debug, Formatter},
+    future::{self, Future},
     sync::Arc,
 };
 
 use hrpc::{
-    client::transport::{self as client_transport, TransportRequest, TransportResponse},
+    client::layer::backoff::Backoff,
+    common::layer::trace::Trace,
     encode::encode_protobuf_message,
     exports::{
         bytes::{Bytes, BytesMut},
+        futures_util::{future::Either, TryFutureExt},
         tower::Service,
     },
+    proto::Error as HrpcError,
+    request::BoxRequest,
+    response::BoxResponse,
     Response,
 };
 use http::Uri;
-use parking_lot::{Mutex, MutexGuard};
 use reqwest::Client as HttpClient;
-use tokio::sync::Mutex as AsyncMutex;
+use std::sync::{Mutex, MutexGuard, RwLock, RwLockReadGuard};
+
+type SpanFnPtr = fn(&BoxRequest) -> Span;
+type OnRequestFnPtr = fn(&BoxRequest, &Span);
+type OnSuccessFnPtr = fn(&BoxResponse, &Span);
+type OnErrorFnPtr = fn(&BoxResponse, &Span, &HrpcError);
+type TraceClient<Transport> =
+    Trace<AddAuth<Transport>, SpanFnPtr, OnRequestFnPtr, OnSuccessFnPtr, OnErrorFnPtr>;
+#[cfg(not(feature = "client_backoff"))]
+type BaseClient<Transport> = TraceClient<Transport>;
+#[cfg(feature = "client_backoff")]
+type BaseClient<Transport> = Backoff<TraceClient<Transport>>;
+type SharedAuthStatus = Arc<RwLock<(AuthStatus, Bytes)>>;
+
+fn add_base_layers<Err, Transport>(
+    transport: Transport,
+    auth_status: SharedAuthStatus,
+) -> BaseClient<Transport>
+where
+    Transport: Service<
+        BoxRequest,
+        Response = BoxResponse,
+        Error = hrpc::client::transport::TransportError<Err>,
+    >,
+    Err: std::error::Error + 'static,
+{
+    let transport = AddAuth {
+        inner: transport,
+        auth_status,
+    };
+
+    let transport = TraceClient::new(
+        transport,
+        |req| tracing::debug_span!("request", endpoint = %req.endpoint(), headers = ?req.header_map()),
+        |_, _| tracing::debug!("processing request"),
+        |_, _| tracing::debug!("request successful"),
+        |_, _, err| tracing::error!("request failed: {}", err),
+    );
+
+    #[cfg(feature = "client_backoff")]
+    let transport = Backoff::new(transport)
+        .clone_extensions_fn(hrpc::client::transport::http::clone_http_extensions);
+
+    transport
+}
 
 #[cfg(feature = "client_web")]
-type GenericClientTransport = client_transport::http::Wasm;
+mod transport {
+    use super::*;
+
+    pub(super) type GenericClientTransport = hrpc::client::transport::http::Wasm;
+    pub(super) type GenericClient = BaseClient<GenericClientTransport>;
+
+    pub(super) fn create_client(
+        homeserver_url: Uri,
+        auth_status: SharedAuthStatus,
+    ) -> ClientResult<GenericClient> {
+        let transport = GenericClientTransport::new(homeserver_url)
+            .map_err(|err| ClientError::Internal(InternalClientError::Transport(err)))?;
+        let transport = add_base_layers(transport, auth_status);
+        Ok(transport)
+    }
+}
+
 #[cfg(all(feature = "client_native", not(feature = "client_web")))]
-type GenericClientTransport = client_transport::http::Hyper;
-type GenericClient = AddAuth<GenericClientTransport>;
+mod transport {
+    use super::*;
+    use hrpc::client::{layer::backoff::Backoff, transport::http};
+
+    pub(super) type GenericClientTransport = http::Hyper;
+    pub(super) type GenericClient = BaseClient<GenericClientTransport>;
+
+    pub(super) fn create_client(
+        homeserver_url: Uri,
+        auth_status: SharedAuthStatus,
+    ) -> ClientResult<GenericClient> {
+        let transport = GenericClientTransport::new(homeserver_url)
+            .map_err(|err| ClientError::Internal(InternalClientError::Transport(err)))?;
+        let transport = add_base_layers(transport, auth_status);
+        Ok(transport)
+    }
+}
+
+use transport::*;
 
 type AuthService = crate::api::auth::auth_service_client::AuthServiceClient<GenericClient>;
 type ChatService = crate::api::chat::chat_service_client::ChatServiceClient<GenericClient>;
@@ -95,19 +178,28 @@ impl AuthStatus {
     }
 }
 
-#[derive(Debug)]
 struct ClientData {
     homeserver_url: Uri,
-    auth_status: Arc<Mutex<(AuthStatus, Bytes)>>,
-    chat: AsyncMutex<ChatService>,
-    auth: AsyncMutex<AuthService>,
-    mediaproxy: AsyncMutex<MediaProxyService>,
-    profile: AsyncMutex<ProfileService>,
-    emote: AsyncMutex<EmoteService>,
-    batch: AsyncMutex<BatchService>,
+    auth_status: SharedAuthStatus,
+    chat: Mutex<ChatService>,
+    auth: Mutex<AuthService>,
+    mediaproxy: Mutex<MediaProxyService>,
+    profile: Mutex<ProfileService>,
+    emote: Mutex<EmoteService>,
+    batch: Mutex<BatchService>,
     http: HttpClient,
 }
 
+impl Debug for ClientData {
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        f.debug_struct("ClientData")
+            .field("homeserver_url", &self.homeserver_url)
+            .field("http", &self.http)
+            .field("auth_status", &self.auth_status)
+            .finish()
+    }
+}
+
 /// Client implementation for Harmony.
 #[derive(Clone, Debug)]
 pub struct Client {
@@ -193,14 +285,9 @@ impl Client {
         let token_bytes = session.session().map_or_else(Bytes::new, |s| {
             Bytes::copy_from_slice(s.session_token.as_bytes())
         });
-        let auth_status = Arc::new(Mutex::new((session, token_bytes)));
+        let auth_status = Arc::new(RwLock::new((session, token_bytes)));
 
-        let transport = GenericClientTransport::new(homeserver_url.clone())
-            .map_err(|err| ClientError::Internal(InternalClientError::Transport(err)))?;
-        let transport = AddAuth {
-            inner: transport,
-            auth_status: auth_status.clone(),
-        };
+        let transport = create_client(homeserver_url.clone(), auth_status.clone())?;
         let inner = hrpc::client::Client::new(transport);
 
         let auth = AuthService::new_inner(inner.clone());
@@ -213,12 +300,12 @@ impl Client {
         let data = ClientData {
             homeserver_url,
             auth_status,
-            chat: AsyncMutex::new(chat),
-            auth: AsyncMutex::new(auth),
-            mediaproxy: AsyncMutex::new(mediaproxy),
-            profile: AsyncMutex::new(profile),
-            emote: AsyncMutex::new(emote),
-            batch: AsyncMutex::new(batch),
+            chat: Mutex::new(chat),
+            auth: Mutex::new(auth),
+            mediaproxy: Mutex::new(mediaproxy),
+            profile: Mutex::new(profile),
+            emote: Mutex::new(emote),
+            batch: Mutex::new(batch),
             http,
         };
 
@@ -227,7 +314,7 @@ impl Client {
         })
     }
 
-    /// Consumes self and starts an event loop with the given handler.
+    /// Starts an event loop with the given handler.
     ///
     /// All socket errors will be logged with tracing. If the handler
     /// function returns `Ok(true)` or `Err(err)`, the function will
@@ -235,17 +322,18 @@ impl Client {
     pub async fn event_loop<'a, Fut, Hndlr>(
         &'a self,
         subs: Vec<EventSource>,
-        handler: Hndlr,
+        mut handler: Hndlr,
     ) -> Result<(), ClientError>
     where
-        Fut: std::future::Future<Output = ClientResult<bool>> + 'a,
-        Hndlr: Fn(&'a Client, crate::api::chat::Event) -> Fut,
+        Fut: Future<Output = ClientResult<bool>>,
+        Hndlr: FnMut(&'a Client, crate::api::chat::Event) -> Fut + Send + 'a,
     {
         let mut sock = self.subscribe_events(subs).await?;
         loop {
             match sock.get_event().await {
                 Ok(Some(ev)) => {
-                    if handler(self, ev).await? {
+                    let fut = handler(self, ev);
+                    if fut.await? {
                         return Ok(());
                     }
                 }
@@ -257,68 +345,76 @@ impl Client {
 
     #[inline(always)]
     /// Get a mutex guard to the chat service.
-    pub async fn chat(&self) -> tokio::sync::MutexGuard<'_, ChatService> {
-        self.data.chat.lock().await
+    pub fn chat(&self) -> MutexGuard<'_, ChatService> {
+        self.data.chat.lock().expect("poisoned")
     }
 
     #[inline(always)]
     /// Get a mutex guard to the auth service.
-    pub async fn auth(&self) -> tokio::sync::MutexGuard<'_, AuthService> {
-        self.data.auth.lock().await
+    pub fn auth(&self) -> MutexGuard<'_, AuthService> {
+        self.data.auth.lock().expect("poisoned")
     }
 
     #[inline(always)]
     /// Get a mutex guard to the mediaproxy service.
-    pub async fn mediaproxy(&self) -> tokio::sync::MutexGuard<'_, MediaProxyService> {
-        self.data.mediaproxy.lock().await
+    pub fn mediaproxy(&self) -> MutexGuard<'_, MediaProxyService> {
+        self.data.mediaproxy.lock().expect("poisoned")
     }
 
     #[inline(always)]
     /// Get a mutex guard to the profile service.
-    pub async fn profile(&self) -> tokio::sync::MutexGuard<'_, ProfileService> {
-        self.data.profile.lock().await
+    pub fn profile(&self) -> MutexGuard<'_, ProfileService> {
+        self.data.profile.lock().expect("poisoned")
     }
 
     #[inline(always)]
     /// Get a mutex guard to the emote service.
-    pub async fn emote(&self) -> tokio::sync::MutexGuard<'_, EmoteService> {
-        self.data.emote.lock().await
+    pub fn emote(&self) -> MutexGuard<'_, EmoteService> {
+        self.data.emote.lock().expect("poisoned")
     }
 
     #[inline(always)]
     /// Get a mutex guard to the batch service.
-    pub async fn batch(&self) -> tokio::sync::MutexGuard<'_, BatchService> {
-        self.data.batch.lock().await
+    pub fn batch(&self) -> MutexGuard<'_, BatchService> {
+        self.data.batch.lock().expect("poisoned")
     }
 
     /// Execute the given request.
-    pub async fn call<Req>(&self, request: Req) -> ClientResult<Req::Response>
+    pub fn call<Req>(
+        &self,
+        request: Req,
+    ) -> impl Future<Output = ClientResult<Req::Response>> + Send + 'static
     where
         Req: Endpoint,
-        Req::Response: prost::Message + Default,
+        Req::Response: prost::Message + Default + 'static,
     {
-        Ok(request.call_with(self).await?.into_message().await?)
+        let fut = request.call_with(self);
+        async move { Ok(fut.await?.into_message().await?) }
     }
 
     /// Execute the given request, but return a [`hrpc::Response`].
-    pub async fn call_response<Req: Endpoint>(
+    pub fn call_response<Req>(
         &self,
         request: Req,
-    ) -> ClientResult<Response<Req::Response>> {
-        request.call_with(self).await
+    ) -> impl Future<Output = ClientResult<Response<Req::Response>>> + Send + 'static
+    where
+        Req: Endpoint,
+        Req::Response: 'static,
+    {
+        request.call_with(self)
     }
 
     /// Execute the given requests a batch (same) request.
     ///
     /// Note that this does not support the convenience types defined in the [`api`] module.
     /// You will need to convert them to the corresponding request type with `Request::from`.
-    pub async fn batch_call<Req>(
+    pub fn batch_call<Req>(
         &self,
         requests: Vec<Req>,
-    ) -> ClientResult<Vec<<Req as Endpoint>::Response>>
+    ) -> impl Future<Output = ClientResult<Vec<<Req as Endpoint>::Response>>> + Send + 'static
     where
         Req: Endpoint + prost::Message,
-        <Req as Endpoint>::Response: prost::Message + Default,
+        <Req as Endpoint>::Response: prost::Message + Default + 'static,
     {
         use prost::Message;
 
@@ -330,18 +426,21 @@ impl Client {
             endpoint: Req::ENDPOINT_PATH.to_string(),
             requests: encoded.collect(),
         };
-        let responses = self.call(batch_req).await?.responses;
-        let mut decoded = Vec::with_capacity(responses.len());
-        for response in responses {
-            let decoded_msg = <Req as Endpoint>::Response::decode(response.as_ref())?;
-            decoded.push(decoded_msg);
+        let fut = self.batch().batch_same(batch_req);
+        async move {
+            let responses = fut.await?.into_message().await?.responses;
+            let mut decoded = Vec::with_capacity(responses.len());
+            for response in responses {
+                let decoded_msg = <Req as Endpoint>::Response::decode(response.as_ref())?;
+                decoded.push(decoded_msg);
+            }
+            Ok(decoded)
         }
-        Ok(decoded)
     }
 
     #[inline(always)]
-    fn auth_status_lock(&self) -> MutexGuard<(AuthStatus, Bytes)> {
-        self.data.auth_status.lock()
+    fn auth_status_lock(&self) -> RwLockReadGuard<(AuthStatus, Bytes)> {
+        self.data.auth_status.read().unwrap()
     }
 
     /// Get the current auth status.
@@ -409,17 +508,15 @@ impl Client {
     /// # Ok(())
     /// # }
     /// ```
-    pub async fn begin_auth(&self) -> ClientResult<()> {
-        let auth_id = self
-            .auth()
-            .await
-            .begin_auth(BeginAuthRequest {})
-            .await?
-            .into_message()
-            .await?
-            .auth_id;
-        self.auth_status_lock().0 = AuthStatus::InProgress(auth_id);
-        Ok(())
+    pub fn begin_auth(&self) -> impl Future<Output = ClientResult<()>> + Send + 'static {
+        let fut = self.auth().begin_auth(BeginAuthRequest {});
+        let auth_status_lock = self.data.auth_status.clone();
+
+        async move {
+            let resp = fut.await?.into_message().await?;
+            auth_status_lock.write().expect("poisoned").0 = AuthStatus::InProgress(resp.auth_id);
+            Ok(())
+        }
     }
 
     /// Request the next authentication step from the server.
@@ -439,34 +536,36 @@ impl Client {
     /// # Ok(())
     /// # }
     /// ```
-    pub async fn next_auth_step(
+    pub fn next_auth_step(
         &self,
         response: AuthStepResponse,
-    ) -> ClientResult<Option<NextStepResponse>> {
+    ) -> impl Future<Output = ClientResult<Option<NextStepResponse>>> + Send + 'static {
         if let AuthStatus::InProgress(auth_id) = self.auth_status() {
-            let step = self
+            let auth_status_lock = self.data.auth_status.clone();
+            let fut = self
                 .auth()
-                .await
-                .next_step(AuthResponse::new(auth_id, response))
-                .await?
-                .into_message()
-                .await?;
+                .next_step(NextStepRequest::new(auth_id, response.into()));
 
-            Ok(
-                if let Some(AuthStep {
+            Either::Left(async move {
+                let step = fut.await?.into_message().await?;
+
+                let step = if let Some(AuthStep {
                     step: Some(auth_step::Step::Session(session)),
                     ..
                 }) = step.step
                 {
                     let token_bytes = Bytes::copy_from_slice(session.session_token.as_bytes());
-                    *self.auth_status_lock() = (AuthStatus::Complete(session), token_bytes);
+                    *auth_status_lock.write().expect("poisoned") =
+                        (AuthStatus::Complete(session), token_bytes);
                     None
                 } else {
                     Some(step)
-                },
-            )
+                };
+
+                Ok(step)
+            })
         } else {
-            Err(ClientError::NoAuthId)
+            Either::Right(future::ready(Err(ClientError::NoAuthId)))
         }
     }
 
@@ -486,17 +585,14 @@ impl Client {
     /// # Ok(())
     /// # }
     /// ```
-    pub async fn prev_auth_step(&self) -> ClientResult<StepBackResponse> {
+    pub fn prev_auth_step(
+        &self,
+    ) -> impl Future<Output = ClientResult<StepBackResponse>> + Send + 'static {
         if let AuthStatus::InProgress(auth_id) = self.auth_status() {
-            Ok(self
-                .auth()
-                .await
-                .step_back(AuthId::new(auth_id))
-                .await?
-                .into_message()
-                .await?)
+            let fut = self.auth().step_back(StepBackRequest::new(auth_id));
+            Either::Left(async move { Ok(fut.await?.into_message().await?) })
         } else {
-            Err(ClientError::NoAuthId)
+            Either::Right(future::ready(Err(ClientError::NoAuthId)))
         }
     }
 
@@ -514,12 +610,15 @@ impl Client {
     /// # Ok(())
     /// # }
     /// ```
-    pub async fn auth_stream(&self) -> ClientResult<AuthSocket> {
+    pub fn auth_stream(&self) -> impl Future<Output = ClientResult<AuthSocket>> + Send + 'static {
         if let AuthStatus::InProgress(auth_id) = self.auth_status() {
-            let inner = self.auth().await.stream_steps(AuthId::new(auth_id)).await?;
-            Ok(AuthSocket { inner })
+            let fut = self.auth().stream_steps(AuthId::new(auth_id));
+            Either::Left(
+                fut.map_ok(|inner| AuthSocket { inner })
+                    .map_err(ClientError::from),
+            )
         } else {
-            Err(ClientError::NoAuthId)
+            Either::Right(future::ready(Err(ClientError::NoAuthId)))
         }
     }
 
@@ -537,20 +636,23 @@ impl Client {
     /// # Ok(())
     /// # }
     /// ```
-    pub async fn subscribe_events(
+    pub fn subscribe_events(
         &self,
         subscriptions: Vec<EventSource>,
-    ) -> ClientResult<EventsSocket> {
-        let (tx, rx) = self.chat().await.stream_events(()).await?.split();
-        let mut socket = EventsSocket {
-            write: EventsWriteSocket { inner: tx },
-            read: EventsReadSocket { inner: rx },
-        };
-        for source in subscriptions {
-            socket.add_source(source).await?;
-        }
+    ) -> impl Future<Output = ClientResult<EventsSocket>> + Send + 'static {
+        let fut = self.chat().stream_events(());
+        async move {
+            let (tx, rx) = fut.await?.split();
+            let mut socket = EventsSocket {
+                write: EventsWriteSocket { inner: tx },
+                read: EventsReadSocket { inner: rx },
+            };
+            for source in subscriptions {
+                socket.add_source(source).await?;
+            }
 
-        Ok(socket)
+            Ok(socket)
+        }
     }
 }
 
@@ -558,12 +660,12 @@ impl Client {
 #[derive(Debug, Clone)]
 pub struct AddAuth<S> {
     inner: S,
-    auth_status: Arc<Mutex<(AuthStatus, Bytes)>>,
+    auth_status: SharedAuthStatus,
 }
 
-impl<S> Service<TransportRequest> for AddAuth<S>
+impl<S> Service<BoxRequest> for AddAuth<S>
 where
-    S: Service<TransportRequest, Response = TransportResponse>,
+    S: Service<BoxRequest>,
 {
     type Response = S::Response;
 
@@ -578,13 +680,9 @@ where
         Service::poll_ready(&mut self.inner, cx)
     }
 
-    fn call(&mut self, mut req: TransportRequest) -> Self::Future {
-        let guard = self.auth_status.lock();
+    fn call(&mut self, mut req: BoxRequest) -> Self::Future {
+        let guard = self.auth_status.read().expect("poisoned");
         if guard.0.is_authenticated() {
-            let req = match &mut req {
-                TransportRequest::Unary(req) => req,
-                TransportRequest::Socket(req) => req,
-            };
             req.get_or_insert_header_map().insert(
                 http::header::AUTHORIZATION,
                 // This is safe on the assumption that servers will never send session tokens
